"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.submitForDefiTxn = exports.claimRewards = exports.registerAddress = exports.isUnclaimedReward = exports.isAddressRegistered = void 0;
const bech32_1 = require("bech32");
const ethersV5_1 = require("ethersV5");
const sign_1 = require("./ledger/sign");
const fs_1 = __importDefault(require("fs"));
const constants_1 = require("./constants");
const flareContractConstants_1 = require("./flareContractConstants");
const utils_1 = require("./utils");
const screenConstants_1 = require("./screenConstants");
const process_1 = require("process");
/**
 * @description checks if the address is registered with the addressBinder contract
 * @param ethAddressToCheck
 * @param network
 */
function isAddressRegistered(ethAddressToCheck, network) {
    return __awaiter(this, void 0, void 0, function* () {
        console.log("Checking Address Registration...");
        const rpcUrl = getRpcUrl(network);
        const addressBinderContractAddress = yield getContractAddress(network, flareContractConstants_1.addressBinderContractName);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, flareContractConstants_1.getAddressBinderABI)();
        const contract = new ethersV5_1.ethers.Contract(addressBinderContractAddress, abi, provider);
        const result = yield contract.cAddressToPAddress(ethAddressToCheck);
        if (result !== '0x0000000000000000000000000000000000000000') {
            console.log(`${constants_1.colorCodes.greenColor}Address associated with key ${ethAddressToCheck}: ${result}${constants_1.colorCodes.resetColor}`);
            return true;
        }
        else {
            console.log(`${constants_1.colorCodes.redColor}No address found for key ${ethAddressToCheck}${constants_1.colorCodes.resetColor}`);
            return false;
        }
    });
}
exports.isAddressRegistered = isAddressRegistered;
/**
 * @description checks if there are any unclaimed rewards with the validatorRewardManager contract
 * @param ethAddressToCheck
 * @param network
 */
function isUnclaimedReward(ethAddressToCheck, network) {
    return __awaiter(this, void 0, void 0, function* () {
        console.log("Checking your Rewards status...");
        const rpcUrl = getRpcUrl(network);
        const validatorRewardManagerContractAddress = yield getContractAddress(network, flareContractConstants_1.validatorRewardManagerContractName);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, flareContractConstants_1.getValidatorRewardManagerABI)();
        const contract = new ethersV5_1.ethers.Contract(validatorRewardManagerContractAddress, abi, provider);
        const rewards = yield contract.getStateOfRewards(ethAddressToCheck);
        const totalRewardNumber = rewards[0];
        const claimedRewardNumber = rewards[1];
        const unclaimedRewards = totalRewardNumber.sub(claimedRewardNumber);
        if (unclaimedRewards.gt(ethersV5_1.BigNumber.from("0"))) {
            console.log(`${constants_1.colorCodes.greenColor}You have unclaimed rewards worth ${unclaimedRewards}${constants_1.colorCodes.resetColor}`);
            return true;
        }
        else {
            console.log(`${constants_1.colorCodes.redColor}No unclaimed rewards found ${constants_1.colorCodes.resetColor}`);
            return false;
        }
    });
}
exports.isUnclaimedReward = isUnclaimedReward;
/**
 * @description cregisters address  with the addressBinder contract
 * @param {RegisterAddressInterface} addressParams
 */
function registerAddress(addressParams) {
    return __awaiter(this, void 0, void 0, function* () {
        const { publicKey, pAddress, cAddress, network, wallet, derivationPath, pvtKey, transactionId } = addressParams;
        const rpcUrl = getRpcUrl(network);
        const addressBinderContractAddress = yield getContractAddress(network, flareContractConstants_1.addressBinderContractName);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, flareContractConstants_1.getAddressBinderABI)();
        const contract = new ethersV5_1.ethers.Contract(addressBinderContractAddress, abi, provider);
        const checksumAddress = ethersV5_1.ethers.utils.getAddress(cAddress);
        const nonce = yield provider.getTransactionCount(checksumAddress);
        const config = (0, constants_1.getConfig)(network);
        const pAddr = (0, utils_1.prefix0x)(Buffer.from(bech32_1.bech32.fromWords(bech32_1.bech32.decode(pAddress.slice(2)).words)).toString('hex'));
        const publicKeyPrefixed = (0, utils_1.prefix0x)(publicKey);
        const gasEstimate = yield contract.estimateGas.registerAddresses(publicKeyPrefixed, pAddr, cAddress, { from: cAddress });
        const gasPrice = yield provider.getGasPrice();
        const populatedTx = yield contract.populateTransaction.registerAddresses(publicKeyPrefixed, pAddr, cAddress);
        const unsignedTx = Object.assign(Object.assign({}, populatedTx), { nonce, chainId: config.networkID, gasPrice, gasLimit: gasEstimate });
        yield signContractTransaction(wallet, unsignedTx, contract, derivationPath, transactionId, pvtKey);
    });
}
exports.registerAddress = registerAddress;
/**
 * @description claims rewards from the ValidatorRewardManager Contract
 * @param {ClaimRewardsInterface} rewardsParams
 */
function claimRewards(rewardsParams) {
    return __awaiter(this, void 0, void 0, function* () {
        const { claimAmount, ownerAddress, receiverAddress, network, wallet, derivationPath, pvtKey, transactionId } = rewardsParams;
        const rewardAmount = ethersV5_1.BigNumber.from(claimAmount).mul(ethersV5_1.ethers.constants.WeiPerEther);
        if (rewardAmount.gt(yield getUnclaimedRewards(ownerAddress, network)) || !rewardAmount.gt(ethersV5_1.BigNumber.from("0"))) {
            throw new Error("Incorrect amount to claim");
        }
        const rpcUrl = getRpcUrl(network);
        const validatorRewardManagerContractAddress = yield getContractAddress(network, flareContractConstants_1.validatorRewardManagerContractName);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, flareContractConstants_1.getValidatorRewardManagerABI)();
        const contract = new ethersV5_1.ethers.Contract(validatorRewardManagerContractAddress, abi, provider);
        const checksumAddress = ethersV5_1.ethers.utils.getAddress(ownerAddress);
        const nonce = yield provider.getTransactionCount(checksumAddress);
        const config = (0, constants_1.getConfig)(network);
        let gasEstimate;
        try {
            gasEstimate = yield contract.estimateGas.claim(ownerAddress, (0, utils_1.prefix0x)(receiverAddress), rewardAmount, false, { from: ownerAddress });
        }
        catch (_a) {
            console.log(`${constants_1.colorCodes.redColor}Incorrect arguments passed${constants_1.colorCodes.resetColor}`);
            (0, process_1.exit)();
        }
        const gasPrice = yield provider.getGasPrice();
        const populatedTx = yield contract.populateTransaction.claim(ownerAddress, (0, utils_1.prefix0x)(receiverAddress), rewardAmount, false);
        const unsignedTx = Object.assign(Object.assign({}, populatedTx), { nonce, chainId: config.networkID, gasPrice, gasLimit: gasEstimate });
        yield signContractTransaction(wallet, unsignedTx, contract, derivationPath, transactionId, pvtKey);
    });
}
exports.claimRewards = claimRewards;
/**
 * @description submits signed ForDefi txn to the chain
 * @param id Id of the txn to submit
 * @param signature signature of the signed txn from ForDefi
 * @param network
 * @returns TxnHash
 */
function submitForDefiTxn(id, signature, network) {
    return __awaiter(this, void 0, void 0, function* () {
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(getRpcUrl(network));
        const serializedSignedTx = ethersV5_1.ethers.utils.serializeTransaction(readUnsignedEVMObject(id), (0, utils_1.prefix0x)(signature));
        const chainId = yield provider.sendTransaction(serializedSignedTx);
        return chainId.hash;
    });
}
exports.submitForDefiTxn = submitForDefiTxn;
function createUnsignedJsonObject(txHash) {
    const signatureRequest = {
        message: txHash.slice(2),
        signer: "",
    };
    const unsignedTxJson = {
        transactionType: flareContractConstants_1.contractTransactionName,
        serialization: "",
        signatureRequests: [signatureRequest],
        unsignedTransactionBuffer: "",
        usedFee: "",
        txDetails: "",
    };
    return unsignedTxJson;
}
function getRpcUrl(network) {
    const config = (0, constants_1.getConfig)(network);
    return `${config.protocol}://${config.ip}/ext/bc/C/rpc`;
}
function getContractAddress(network, contractName) {
    var _a;
    return __awaiter(this, void 0, void 0, function* () {
        const rpcUrl = getRpcUrl(network);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, flareContractConstants_1.getFlareContractRegistryABI)();
        if (network != "flare" && network != "costwo")
            throw new Error("Invalid network passed");
        const contract = new ethersV5_1.ethers.Contract(flareContractConstants_1.defaultContractAddresses.flareContractRegistryAddress[network], abi, provider);
        const result = yield contract.getContractAddressByName(contractName);
        if (result !== '0x0000000000000000000000000000000000000000')
            return result;
        const defaultAddress = (_a = flareContractConstants_1.defaultContractAddresses[contractName]) === null || _a === void 0 ? void 0 : _a[network];
        if (defaultAddress)
            return defaultAddress;
        throw new Error("Contract Address not found");
    });
}
function readUnsignedEVMObject(id) {
    const fname = `${constants_1.forDefiDirectory}/${constants_1.forDefiUnsignedTxnDirectory}/${id}.unsignedEVMObject.json`;
    if (!fs_1.default.existsSync(fname)) {
        throw new Error(`unsigned EVM Object file ${fname} does not exist`);
    }
    const serialization = fs_1.default.readFileSync(fname).toString();
    let file = JSON.parse(serialization);
    file.gasPrice = ethersV5_1.BigNumber.from(file.gasPrice.hex);
    file.gasLimit = ethersV5_1.BigNumber.from(file.gasLimit.hex);
    return file;
}
function saveUnsignedEVMObject(unsignedTx, id) {
    const fname = `${constants_1.forDefiDirectory}/${constants_1.forDefiUnsignedTxnDirectory}/${id}.unsignedEVMObject.json`;
    if (fs_1.default.existsSync(fname)) {
        throw new Error(`unsignedTx file ${fname} already exists`);
    }
    const serialization = JSON.stringify(unsignedTx, null, 2);
    fs_1.default.mkdirSync(`${constants_1.forDefiDirectory}/${constants_1.forDefiUnsignedTxnDirectory}`, { recursive: true });
    fs_1.default.writeFileSync(fname, serialization);
}
function getUnclaimedRewards(ethAddress, network) {
    return __awaiter(this, void 0, void 0, function* () {
        const rpcUrl = getRpcUrl(network);
        const validatorRewardManagerContractAddress = yield getContractAddress(network, flareContractConstants_1.validatorRewardManagerContractName);
        const provider = new ethersV5_1.ethers.providers.JsonRpcProvider(rpcUrl);
        const abi = (0, flareContractConstants_1.getValidatorRewardManagerABI)();
        const contract = new ethersV5_1.ethers.Contract(validatorRewardManagerContractAddress, abi, provider);
        const rewards = yield contract.getStateOfRewards(ethAddress);
        const totalRewardNumber = rewards[0];
        const claimedRewardNumber = rewards[1];
        const unclaimedRewards = totalRewardNumber.sub(claimedRewardNumber);
        return unclaimedRewards;
    });
}
function signContractTransaction(wallet, unsignedTx, contract, derivationPath, transactionId, pvtKey) {
    return __awaiter(this, void 0, void 0, function* () {
        const serializedUnsignedTx = ethersV5_1.ethers.utils.serializeTransaction(unsignedTx);
        const txHash = ethersV5_1.ethers.utils.keccak256(serializedUnsignedTx);
        const unsignedTxObj = createUnsignedJsonObject(txHash);
        if (wallet === Object.keys(screenConstants_1.walletConstants)[0]) {
            if (!derivationPath)
                throw new Error("No derivation path passed");
            const sign = yield (0, sign_1.ledgerSign)(unsignedTxObj, derivationPath);
            const serializedSignedTx = ethersV5_1.ethers.utils.serializeTransaction(unsignedTx, (0, utils_1.prefix0x)(sign.signature));
            console.log("Submitting txn to the chain");
            yield contract.provider.sendTransaction(serializedSignedTx);
        }
        else if (wallet === Object.keys(screenConstants_1.walletConstants)[1]) {
            if (!transactionId)
                throw new Error("No transaction Id passed");
            (0, utils_1.saveUnsignedTxJson)(unsignedTxObj, transactionId);
            saveUnsignedEVMObject(unsignedTx, transactionId);
        }
        else if (wallet === Object.keys(screenConstants_1.walletConstants)[2]) {
            if (!pvtKey)
                throw new Error("No private key passed");
            const wallet = new ethersV5_1.ethers.Wallet(pvtKey);
            const signedTx = yield wallet.signTransaction(unsignedTx);
            yield contract.provider.sendTransaction(signedTx);
        }
    });
}
